local _, ns = ...
local OnCast = ns.addon
local Console = OnCast:NewModule("Console")

-- Lua APIs
local typeof = type
local sformat = string.format
local supper = string.upper
local slower = string.lower
local sgsub = string.gsub
local tonumber = tonumber
local tostring = tostring

-- WoW APIs
local WOW = OnCast.WOWAPI or {}
WOW.GetSpellInfo = GetSpellInfo  -- https://wow.gamepedia.com/API_GetSpellInfo
WOW.GetSpellNameAndId = function(spell) 
    local spellName,_,_,_,_,_,spellId = WOW.GetSpellInfo(spell)
    return spellName, spellId
end

-- Modules
local Util = nil
local Event = nil

-- Constants
local TYPE_SPLIT = ','
local CONDITION_SPLIT = ','
local VAR_SPLIT = ';'
local NAME_VAR_SPLIT = "%(%)$"

-- Module functions
function Console:OnInitialize()
    -- Modules
    Util = OnCast:GetModule("Util")
    Event = OnCast:GetModule("Event")

    Util.addSplitPattern(TYPE_SPLIT)
    Util.addSplitPattern(CONDITION_SPLIT)
    Util.addSplitPattern(VAR_SPLIT)
    Util.addSplitPattern(NAME_VAR_SPLIT)
end

function Console:OnEnable()
end

function Console:OnDisable()
end

-- Temporary arrays for parsing
local _tempArray1 = {}
local _tempArray2 = {}

-- Validate the argurment as not nil and less than the length of args (AceConsole appends a final value)
local function argValid(args, idx)
    return (args[idx] and idx < #args)
end

-- System command for /oncast
function Console.parseSystemCommand(sysCmd, sysSub)
    if not sysCmd then return false end
    sysCmd = supper(sysCmd)

    -- Print version
    if sysCmd == "VERSION" or sysCmd == "-V" then
        OnCast.printInfoMessage(sformat("Version: %d.%d.%d", OnCast.VERSION.MAJOR, OnCast.VERSION.MINOR, OnCast.VERSION.PATCH))
        return true
    end

    -- Print Usage Help
    if sysCmd == "HELP" or sysCmd == "-H" or sysCmd == "?" then
        local topic = sysSub and supper(sysSub)
        if topic then
            if topic == "TOKEN" or topic == "TOKENS" then
                OnCast.printInfoMessage("-- Message Tokens --", OnCast.HEADING_COLOR)
                OnCast.printInfoMessage("$spell $sp: The name of the spell being cast")
                OnCast.printInfoMessage("$target $t: The name of the target")
                OnCast.printInfoMessage("$rt: The raid target marker of the specified @unit")
                OnCast.printInfoMessage("$event $ev: The current EVENT of the cast or spell")
                OnCast.printInfoMessage("$outcome $oc: The outcome of the cast or spell")
                OnCast.printInfoMessage("$reason $rsn: The reason the cast FAILED")
                OnCast.printInfoMessage("$delay $dl: The current total delay of the cast")
                OnCast.printInfoMessage("$time $tm: The current or final cast time")
                OnCast.printInfoMessage("$group $grp: The raid group of the target")
                OnCast.printInfoMessage("$amount $amt: Damage, Healing, Power Gain or Aura stacks")
                OnCast.printInfoMessage("$xspell $xsp: The target's spell that was affected")
                OnCast.printInfoMessage("$xtype $xtp: \"Buff\" \"Debuff\" or \"Mana\" \"Rage\" etc.")
                return true
            end
            if topic == "CONDITION" or topic == "CONDITIONS" then
                OnCast:GetModule("MacroConditionals").Help()
                return true
            end
            if topic == "EVENT" or topic == "EVENTS" then
                OnCast.printInfoMessage("-- Event Aliases --", OnCast.HEADING_COLOR)
                OnCast.printInfoMessage("START - STARTED  SR")
                OnCast.printInfoMessage("STOP - STOPPED  SP")
                OnCast.printInfoMessage("SUCCEEDED - SUCCESS  SU")
                OnCast.printInfoMessage("FAILED - FAILURE  FAIL  FL")
                OnCast.printInfoMessage("INTERRUPTED - INTERRUPT  IN")
                OnCast.printInfoMessage("HIT - HT")
                OnCast.printInfoMessage("EFFECT - EF")
                OnCast.printInfoMessage("EXTRA - EX")
                OnCast.printInfoMessage("MISS - MS")
                return true
            end
            if topic == "EXEVENT" or topic == "EXEVENTS" then
                OnCast.printInfoMessage("-- Extended Events --", OnCast.HEADING_COLOR)
                OnCast.printInfoMessage("HIT(crit;minimum): Direct damage or healing.")
                OnCast.printInfoMessage(" - Crit: Only show critical hits")
                OnCast.printInfoMessage(" - Minimum: Minimum amount to show")
                OnCast.printInfoMessage("EFFECT: Special effects such as buff, interrupt, dispel, etc.")
                OnCast.printInfoMessage("EXTRA(Effect_Name): An extra effect")
                OnCast.printInfoMessage(" - Effect_Name: (required) The name of the effect")
                OnCast.printInfoMessage("MISS: Resisted, Reflected, Dodged, Immune, etc.")
                return true
            end
            if topic == "CHANNEL" or topic == "CHANNELS" then
                OnCast.printInfoMessage("-- Channel Aliases --", OnCast.HEADING_COLOR)
                OnCast.printInfoMessage("RAID - R")
                OnCast.printInfoMessage("RAID10 - R10")
                OnCast.printInfoMessage("RAID20 - R20")
                OnCast.printInfoMessage("RAIDPARTY - RPARTY  RP")
                OnCast.printInfoMessage("PARTY - P")
                OnCast.printInfoMessage("BATTLEGROUND - BG")
                OnCast.printInfoMessage("ARENA - AR")
                OnCast.printInfoMessage("WHISPER - W")
                OnCast.printInfoMessage("SAY - S")
                OnCast.printInfoMessage("EMOTE - E")
                return true
            end
            local extension = OnCast.EXTENDED_CHANNELS[topic]
            if extension then
                extension.Help()
                return true
            end
            OnCast.printInfoMessage(sformat("Unknown Help Topic '%s'", topic))
            OnCast.printInfoMessage(sformat("Topics: token, condition, event, exevent, channel, %s", slower(Util.concatTableKeys(OnCast.EXTENDED_CHANNELS, ", "))))
            return true
        end
        OnCast.printInfoMessage("-- Usage: /oncast [conditions] events channels \"Message\"", OnCast.HEADING_COLOR)
        OnCast.printInfoMessage("condition: @unit btn:n mod:key. Keys: shift  ctrl  alt")
        OnCast.printInfoMessage("event: START  STOP  SUCCEEDED  FAILED  INTERRUPTED  DELAYED")
        OnCast.printInfoMessage("exEvent: HIT  EFFECT  EXTRA  MISS")
        OnCast.printInfoMessage("channel: RAID  PARTY  BATTLEGROUND  SAY  WHISPER  EMOTE")
        OnCast.printInfoMessage("spChannel: RAID10  RAID20  RAIDPARTY")
        OnCast.printInfoMessage(sformat("exChannel : %s", Util.concatTableKeys(OnCast.EXTENDED_CHANNELS, "  ")))
        OnCast.printInfoMessage("\"Message\": Message to display enclosed in \" \"")
        OnCast.printInfoMessage("-- For more info use /oncast help topic")
        OnCast.printInfoMessage(sformat("Topics: token, condition, event, exevent, channel, %s", slower(Util.concatTableKeys(OnCast.EXTENDED_CHANNELS, ", "))))
        return true
    end

    return false
end

-- Cache for macro events and channels
local eventCache = {}
local channelCache = {}

-- Command for /oncast
function Console.parseCommand(cmd, argEvents, argChannels, argMessage)
    -- print(argEvents, argChannels, argMessage)
    local n = 0

    -- Check for event
    if not argEvents then
        OnCast.printErrorMessage("Expected: event")
        OnCast.printInfoMessage("Type '/oncast help' for further details")
        return false
    end

    -- Events
    local cachedEvents = eventCache[argEvents]
    if cachedEvents then
        cmd.events = cachedEvents.events
        cmd.eventVars = cachedEvents.vars
        cmd.meta.hasExtendedEvents = cachedEvents.hasExtendedEvents
    else
        cmd.events = {}
        cmd.eventVars = {}
        local hasExtendedEvents = false
        n = Util.stringSplitEF(argEvents, TYPE_SPLIT, _tempArray1)
        for i=1, n do
            -- Reset tempArray2
            _tempArray2[1] = nil
            _tempArray2[2] = nil
            -- Split event name from event vars
            Util.stringSplitEF(_tempArray1[i], NAME_VAR_SPLIT, _tempArray2) 
            local inputName = supper(_tempArray2[1])
            -- Check standard events
            local eventName = OnCast.SUPPORTED_EVENTS[inputName]
            if eventName then
                cmd.events[i] = eventName
            else -- Check extended events
                eventName = OnCast.SUPPORTED_EXTENDED_EVENTS[inputName]
                if eventName then
                    local exEvent = OnCast.EXTENDED_EVENTS[eventName]
                    local vars, nvars = Util.stringSplit(_tempArray2[2], VAR_SPLIT)
                    local ok, eventVars = exEvent.VarsToTable(vars, nvars)
                    -- Check vars parsed OK
                    if not ok then return false end
                    -- Set name and vars
                    cmd.events[i] = eventName
                    cmd.eventVars[i] = eventVars
                    hasExtendedEvents = true
                end
            end
            -- Not Supported
            if not cmd.events[i] then
                OnCast.printErrorMessage(sformat("Unsupported event: %s", inputName))
                OnCast.printInfoMessage("Type '/oncast help' for further details")
                return false
            end
        end

        -- Set meta data
        cmd.meta.hasExtendedEvents = hasExtendedEvents

        -- Create cached events
        local cachedEvents = {}
        cachedEvents.events = cmd.events
        cachedEvents.vars = cmd.eventVars
        cachedEvents.hasExtendedEvents = hasExtendedEvents
        eventCache[argEvents] = cachedEvents
    end

    -- Check for channels
    if not argChannels then
        OnCast.printErrorMessage("Expected: channel")
        OnCast.printInfoMessage("Type '/oncast help' for further details")
        return false
    end

    -- Channels
    local cachedChannels = channelCache[argChannels]
    if cachedChannels then
        cmd.channels = cachedChannels.channels
        cmd.channelVars = cachedChannels.vars
        cmd.meta.skipExtended = cachedChannels.skipExtended
        cmd.meta.skipStandard = cachedChannels.skipStandard
    else
        cmd.channels = {}
        cmd.channelVars = {}
        local skipStandard, skipExtended = true, true
        n = Util.stringSplitEF(argChannels, TYPE_SPLIT, _tempArray1)
        for i=1, n do
            -- Reset temp
            _tempArray2[1] = nil
            _tempArray2[2] = nil
            -- Split channel name from channel vars
            Util.stringSplitEF(_tempArray1[i], NAME_VAR_SPLIT, _tempArray2) 
            local inputName = supper(_tempArray2[1])
            -- Check standard channels
            local channelName = OnCast.STANDARD_CHANNELS[inputName]
            if channelName then
                cmd.channels[i] = channelName
                skipStandard = false
            else -- Check extended channels
                local exChannel = OnCast.EXTENDED_CHANNELS[inputName]
                if exChannel then                
                    local vars, nvars = Util.stringSplit(_tempArray2[2], VAR_SPLIT)
                    local ok, channelVars = exChannel.VarsToTable(vars, nvars)
                    -- Check vars parsed OK
                    if not ok then return false end
                    -- Set name and vars
                    cmd.channels[i] = inputName
                    cmd.channelVars[i] = channelVars
                    skipExtended = false
                end
            end
            -- Not supported
            if not cmd.channels[i] then
                OnCast.printErrorMessage(sformat("Unsupported channel: %s", inputName))
                OnCast.printInfoMessage("Type '/oncast help' for further details")
                return false
            end
        end
        
        -- Set meta data
        cmd.meta.skipStandard = skipStandard
        cmd.meta.skipExtended = skipExtended

        -- Create cached channel
        local cachedChannels = {}
        cachedChannels.channels = cmd.channels
        cachedChannels.vars = cmd.channelVars
        cachedChannels.skipStandard = cmd.meta.skipStandard
        cachedChannels.skipExtended = cmd.meta.skipExtended
        channelCache[argChannels] = cachedChannels
    end

    -- Check for message
    if not argMessage then
        OnCast.printErrorMessage("Expected: \"Message\"")
        OnCast.printInfoMessage("Type '/oncast help' for further details")
        return false
    end

   -- Message
    cmd.message = argMessage

    return true
end

-- System command of /oncast-spell
function Console.parseSpellSystemCommand(args)
    local sysCmd = argValid(args, 1) and supper(args[1])
    if not sysCmd then return false end
    
    -- Print version
    if sysCmd == "VERSION" or sysCmd == "-V" then
        OnCast.printInfoMessage(sformat("Version: %d.%d.%d", OnCast.VERSION.MAJOR, OnCast.VERSION.MINOR, OnCast.VERSION.PATCH))
        return true
    end

    -- Print Usage Help
    if sysCmd == "HELP" or sysCmd == "-H" or sysCmd == "?" then
        OnCast.printInfoMessage("-- Usage: /oncast-spell [remove|show] \"Spell\" events channels \"Message\"", OnCast.HEADING_COLOR)
        OnCast.printInfoMessage("remove: Remove events or channels for the Spell")
        OnCast.printInfoMessage("show: Shows events for the Spell")
        OnCast.printInfoMessage("\"Spell\": The Name of the Spell enclosed in \" \", or the SpellId")
        OnCast.printInfoMessage("event: START  STOP  SUCCEEDED  FAILED  INTERRUPTED  DELAYED")
        OnCast.printInfoMessage("exEvent: HIT  EFFECT  EXTRA  MISS")
        OnCast.printInfoMessage("channel: RAID  PARTY  BATTLEGROUND  SAY  WHISPER  EMOTE")
        OnCast.printInfoMessage("spChannel: RAID10  RAID20  RAIDPARTY")
        OnCast.printInfoMessage(sformat("exChannel: %s", Util.concatTableKeys(OnCast.EXTENDED_CHANNELS, "  ")))
        OnCast.printInfoMessage("\"Message\": Message to display enclosed in \" \"")
        OnCast.printInfoMessage("-- For more info use /oncast help topic")
        OnCast.printInfoMessage(sformat("Topics: token, event, %s", slower(Util.concatTableKeys(OnCast.EXTENDED_CHANNELS, ", "))))
        return true
    end
    
    if sysCmd == "DEBUG" or sysCmd == "-D" then
        OnCast.printErrorMessage("Debug has been moved to /oncast-debug")
        return true
    end

    return false
end

-- Command for /oncast-spell
function Console.parseSpellCommand(cmd, args)
    --local args = {...} -- Argument list
    --for i = 0, #args do print(i, args[i]) end -- Debug
    
    local idx = 1
    local n = 0
    
    -- Check for remove, show or spell
    if not argValid(args, idx) then
        OnCast.printErrorMessage("Expected: remove, show or \"Spell\"")
        OnCast.printInfoMessage("Type '/oncast-spell help' for further details")
        return false
    end

    -- Modifiers
    local mod = supper(args[idx])
    if mod == "REMOVE" then
        cmd.remove = true
        idx = idx + 1
    elseif mod == "SHOW" then
        cmd.show = true
        idx = idx + 1
    end
    
    -- Check for spell
    if not argValid(args, idx) then
        -- End here for show
        if cmd.show then return true end

        OnCast.printErrorMessage("Expected: \"Spell\"")
        OnCast.printInfoMessage("Type '/oncast-spell help' for further details")
        return false
    end

    -- Spell
    local spell = tonumber(args[idx]) or sgsub(args[idx], "_", " ") -- As an id (number) or name (string)
    local spellName,spellId = WOW.GetSpellNameAndId(spell)
    if not (spellName and spellId) then
        OnCast.printErrorMessage(sformat("Unknown Spell: %s", tostring(spell)))
        return false
    end
    cmd.spell = typeof(spell) == "string" and spellName or spellId
    -- Check if ignored
    if (not (cmd.remove or cmd.show)) and Event.IGNORE_SPELL[spellId] then
        OnCast.printErrorMessage(sformat("Unable to Register Spell: %s (%s) is on Ignore list", spellName, tostring(spellId)))
        return false
    end
    idx = idx + 1

    -- Check for event
    if not argValid(args, idx) then
        -- End here for remove
        if cmd.remove then return true end
        -- End here for show
        if cmd.show then return true end

        OnCast.printErrorMessage("Expected: events")
        OnCast.printInfoMessage("Type '/oncast-spell help' for further details")
        return false
    end

    -- Events
    cmd.events = {}
    cmd.eventVars = {}
    local hasExtendedEvents = false
    n = Util.stringSplitEF(args[idx], TYPE_SPLIT, _tempArray1)
    for i=1, n do
        -- Reset tempArray2
        _tempArray2[1] = nil
        _tempArray2[2] = nil
        -- Split event name from event vars
        Util.stringSplitEF(_tempArray1[i], NAME_VAR_SPLIT, _tempArray2) 
        local inputName = supper(_tempArray2[1])
        -- Check standard events
        local eventName = OnCast.SUPPORTED_EVENTS[inputName]
        if eventName then
            cmd.events[i] = eventName
        else -- Check extended events
            eventName = OnCast.SUPPORTED_EXTENDED_EVENTS[inputName]
            if eventName then
                local exEvent = OnCast.EXTENDED_EVENTS[eventName]
                local vars, nvars = Util.stringSplit(_tempArray2[2], VAR_SPLIT)
                local ok, eventVars = exEvent.VarsToTable(vars, nvars)
                -- Check vars parsed OK
                if not ok then return false end
                -- Set name and vars
                cmd.events[i] = eventName
                cmd.eventVars[i] = eventVars
                hasExtendedEvents = true
            end
        end
        -- Not Supported
        if not cmd.events[i] then
            OnCast.printErrorMessage(sformat("Unsupported event: %s", inputName))
            OnCast.printInfoMessage("Type '/oncast help' for further details")
            return false
        end
        -- Set meta data
        cmd.meta.hasExtendedEvents = hasExtendedEvents
    end
    idx = idx + 1

    -- End here for show
    if cmd.show then return true end

    -- Check for channels
    if not argValid(args, idx) then
        -- End here for remove
        if cmd.remove then return true end

        OnCast.printErrorMessage("Expected: channels")
        OnCast.printInfoMessage("Type '/oncast-spell help' for further details")
        return false
    end

    -- Channels
    cmd.channels = {}
    cmd.channelVars = {}
    local skipStandard, skipExtended = true, true
    n = Util.stringSplitEF(args[idx], TYPE_SPLIT, _tempArray1)
    for i=1, n do
        -- Reset temp
        _tempArray2[1] = nil
        _tempArray2[2] = nil
        -- Split channel name from channel vars
        Util.stringSplitEF(_tempArray1[i], NAME_VAR_SPLIT, _tempArray2) 
        local inputName = supper(_tempArray2[1])
        -- Check standard channels
        local channelName = OnCast.STANDARD_CHANNELS[inputName]
        if channelName then
            cmd.channels[i] = channelName
            skipStandard = false
        else -- Check extended channels
            local exChannel = OnCast.EXTENDED_CHANNELS[inputName]
            if exChannel then                
                local vars, nvars = Util.stringSplit(_tempArray2[2], VAR_SPLIT)
                local ok, channelVars = exChannel.VarsToTable(vars, nvars, true)
                -- Check vars parsed OK
                if not ok then return false end
                -- Set name and vars
                cmd.channels[i] = inputName
                cmd.channelVars[i] = channelVars
                skipExtended = false
            end
        end
        -- Not supported
        if not cmd.channels[i] then
            if cmd.remove then -- Take name as is for remove
                cmd.channels[i] = inputName
            else
                OnCast.printErrorMessage(sformat("Unsupported channel: %s", inputName))
                OnCast.printInfoMessage("Type '/oncast help' for further details")
                return false
            end
        end
    end
    idx = idx + 1
    -- Set meta data
    cmd.meta.skipStandard = skipStandard
    cmd.meta.skipExtended = skipExtended
    
    -- End here for remove
    if cmd.remove then return true end

    -- Check for message
    if not argValid(args, idx) then
        OnCast.printErrorMessage("Expected: \"Message\"")
        OnCast.printInfoMessage("Type '/oncast-spell help' for further details")
        return false
    end

   -- Message
    cmd.message = args[idx]

    return true
end
