local _, ns = ...
local OnCast = ns.addon
local ExtEvent = OnCast:NewModule("ExtEvent", "AceEvent-3.0")

-- Lua APIs
local select = select
--local mceil = math.ceil
--local mfmod = math.fmod
local sformat = string.format
local supper = string.upper
local slower = string.lower
local sgsub = string.gsub
--local smatch = string.match
local tostring = tostring
local tonumber = tonumber

-- WoW APIs
local WOW = OnCast.WOWAPI or {}
WOW.GetTime = GetTime
WOW.GetSpellInfo = GetSpellInfo  -- https://wow.gamepedia.com/API_GetSpellInfo
WOW.GetCurrentEventInfo = CombatLogGetCurrentEventInfo -- https://wow.gamepedia.com/API_CombatLogGetCurrentEventInfo
WOW.GetSpellCooldown = GetSpellCooldown -- https://wow.gamepedia.com/API_GetSpellCooldown
WOW.GetSpellLossOfControlCooldown = GetSpellLossOfControlCooldown -- https://wow.gamepedia.com/API_GetSpellLossOfControlCooldown

-- Modules
local Util = nil

-- Extended Event Types
local EVENT_TYPE = {
    HIT = "HIT",
    EFFECT = "EFFECT",
    EXTRA = "EXTRA",
    MISS = "MISS",
    -- Special Events
    FAILED_COOLDOWN = "FAILED_COOLDOWN",
    FAILED_LOSSOFCONTROL = "FAILED_LOSSOFCONTROL"
}

local SUPPORTED_EXTENDED_EVENTS = {
    ["HIT"] = EVENT_TYPE.HIT,
    ["HT"] = EVENT_TYPE.HIT,
    ["EFFECT"] = EVENT_TYPE.EFFECT,
    ["EF"] = EVENT_TYPE.EFFECT,
    ["EXTRA"] = EVENT_TYPE.EXTRA,
    ["EX"] = EVENT_TYPE.EXTRA,
    ["MISS"] = EVENT_TYPE.MISS,
    ["MS"] = EVENT_TYPE.MISS,
    ["FAILED_COOLDOWN"] = EVENT_TYPE.FAILED_COOLDOWN,
    ["FAIL_CD"] = EVENT_TYPE.FAILED_COOLDOWN,
    ["FCD"] = EVENT_TYPE.FAILED_COOLDOWN,
    ["FAILED_LOSSOFCONTROL"] = EVENT_TYPE.FAILED_LOSSOFCONTROL,
    ["FAIL_LOC"] = EVENT_TYPE.FAILED_LOSSOFCONTROL,
    ["FLC"] = EVENT_TYPE.FAILED_LOSSOFCONTROL
}
OnCast.SUPPORTED_EXTENDED_EVENTS = SUPPORTED_EXTENDED_EVENTS

local function isBlank(value)
    return (not value) or value == "_"
end

-- Extended Outcome Events
local EXTENDED_EVENTS = { }
-- HIT
local IS_CRIT = {["CRITICAL"] = true, ["CRIT"] = true, ["C"] = true, ["TRUE"] = true, ["T"] = true}
EXTENDED_EVENTS[EVENT_TYPE.HIT] = {
    VarsToTable = function(vars, n)
        if n <= 0 then return true,nil end
        -- Crit
        if not isBlank(vars[1]) then
            vars.critOnly = IS_CRIT[supper(vars[1])]
        end
        vars[1] = nil

        -- Minimum
        if not isBlank(vars[2]) then
            vars.minimum = tonumber(vars[2]) or 0
        end
        vars[2] = nil

        -- Clear any other values
        for i=3, n do vars[i] = nil end

        return true, vars
    end,
    VarsToArray = function(vars) 
        local array = {}
        array[1] = vars.critOnly and "Crit" or "All"
        array[2] = vars.minimum and tostring(vars.minimum) or "_"
        return array
    end
}
-- EFFECT
EXTENDED_EVENTS[EVENT_TYPE.EFFECT] = {
    VarsToTable = function(vars, n) return true, nil end,
    VarsToArray = function(vars) return {} end
}
-- EXTRA
EXTENDED_EVENTS[EVENT_TYPE.EXTRA] = {
    VarsToTable = function(vars, n)
        -- Effect Name
        local effectName = vars[1]
        if isBlank(effectName) then
            OnCast.printErrorMessage("EXTRA Effect Name Required!")
            return false, nil
        end
        local spellId = tonumber(effectName)
        if spellId then
            effectName = WOW.GetSpellInfo(spellId)
            if not effectName then
                OnCast.printErrorMessage(sformat("Unknown Spell Id %d", spellId))
                return false, nil
            end
        else
            -- Replace _ (underscroe) with  (space)
            effectName = sgsub(effectName, "_", " ")
        end
        
        vars.effectName = effectName
        vars[1] = nil

        -- Clear any other values
        for i=2, n do vars[i] = nil end

        return true, vars
    end,
    VarsToArray = function(vars) 
        local array = {}
        array[1] = vars.effectName or "_"
        return array
    end
}
-- MISS
EXTENDED_EVENTS[EVENT_TYPE.MISS] = {
    VarsToTable = function(vars, n) return true, nil end,
    VarsToArray = function(vars) return {} end
}
OnCast.EXTENDED_EVENTS = EXTENDED_EVENTS

-- FAIL_COOLDOWN
EXTENDED_EVENTS[EVENT_TYPE.FAILED_COOLDOWN] = {
    VarsToTable = function(vars, n)
        if n <= 0 then return true,nil end
        -- Minimum
        if not isBlank(vars[1]) then
            vars.minimum = tonumber(vars[1]) or nil
        end
        vars[1] = nil

        -- Maximum
        if not isBlank(vars[2]) then
            vars.maximum = tonumber(vars[2]) or nil
        end
        vars[2] = nil

        -- Clear any other values
        for i=3, n do vars[i] = nil end

        return true, vars
    end,
    VarsToArray = function(vars)
        local array = {}
        array[1] = vars.minimum and tostring(vars.minimum) or "_"
        array[2] = vars.maximum and tostring(vars.maximum) or "_"
        return array
    end
}

-- FAIL_LOSSCONTROL
EXTENDED_EVENTS[EVENT_TYPE.FAILED_LOSSOFCONTROL] = {
    VarsToTable = function(vars, n)
        if n <= 0 then return true,nil end
        -- Minimum
        if not isBlank(vars[1]) then
            vars.minimum = tonumber(vars[1]) or nil
        end
        vars[1] = nil

        -- Maximum
        if not isBlank(vars[2]) then
            vars.maximum = tonumber(vars[2]) or nil
        end
        vars[2] = nil

        -- Clear any other values
        for i=3, n do vars[i] = nil end

        return true, vars
    end,
    VarsToArray = function(vars)
        local array = {}
        array[1] = vars.minimum and tostring(vars.minimum) or "_"
        array[2] = vars.maximum and tostring(vars.maximum) or "_"
        return array
    end
}

-- Values
local PlayerGUID = UnitGUID("player")

-- Module functions
function ExtEvent:OnInitialize()
    Util = OnCast:GetModule("Util")
end

function ExtEvent:OnEnable()
    -- Subscribe to events
    ExtEvent:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", ExtEvent.OnCombatLogEvent)
end

function ExtEvent:OnDisable()
    -- Unsubscribe all events
    ExtEvent:UnregisterAllEvents()
end

local DEBUG_BLACKLIST = {
    ["Auto Shot"] = true,
    ["Shoot"] = true, -- Wand
    ["Shoot Bow"] = true,
    ["Shoot Gun"] = true,
    ["Shoot Crossbow"] = true,
    ["Throw"] = true,
}

-- List of events to monitor
local _MonitoredEvents = {}

local function HandleCombatLogEvent(eventName, ...)
    -- Standard combat log event parameters
    local timestamp,subEvent,_,sourceGUID,sName,sFlags,srFlags,targetGUID,targetName,tFlags,trFlags = ...
    -- Only take events from the player
    if not (sourceGUID == PlayerGUID) then return end

    -- Check monitored event handlers
    local eventHandlerFunc = _MonitoredEvents[subEvent]
    if eventHandlerFunc then
        eventHandlerFunc(timestamp, subEvent, targetGUID, targetName, select(12,...))
    end
end

function ExtEvent.OnCombatLogEvent(eventName)
    HandleCombatLogEvent(eventName, WOW.GetCurrentEventInfo())
end

-- List of Spells to monitor for extended outcome
local _MonitoredSpells = {}
local copyExclude = {["announcement"] = true}
-- Monitor a spell for the given outcome
function ExtEvent.MonitorSpell(announcement, castData)
    local spell = {}
    local cast = Util.tableDeepCopy(castData, copyExclude) -- Copy the basic cast data

    -- Extra
    local onExtra = announcement[EVENT_TYPE.EXTRA]
    if onExtra and onExtra.vars and onExtra.vars.effectName then
        local effectName = onExtra.vars.effectName
        onExtra.cast = Util.tableDeepCopy(cast) -- Create seperate cast data
        onExtra.cast.spell.name = effectName
        onExtra.name = effectName
        spell.extraEffect = onExtra
    end

    -- Hit
    local onHit = announcement[EVENT_TYPE.HIT]
    if onHit then
        onHit.cast = cast
        spell[EVENT_TYPE.HIT] = onHit
    end
    
    -- Effect
    local onEffect = announcement[EVENT_TYPE.EFFECT]
    if onEffect then
        onEffect.cast = cast
        spell[EVENT_TYPE.EFFECT] = onEffect
    end

    -- Miss
    local onMiss = announcement[EVENT_TYPE.MISS]
    if onMiss then
        onMiss.cast = cast
        spell[EVENT_TYPE.MISS] = onMiss
    end

    _MonitoredSpells[castData.spell.name] = spell
end

-- List of spells that should set extra effects of success
local EXTRA_ON_SUCCESS = {
    ["Counterspell"] = true, -- Only produces a base effect if a spell is Interrupted, Improved can still apply silence
}

-- List of succeeded casts waiting for an outcome
local _SucceededSpells = {}

-- Set the extra effect of a spell
local function setExtraEffect(spell, timestamp)
    if spell.extraEffect then
        local onEffect = spell.extraEffect
        _SucceededSpells[onEffect.name] = {
            timestamp = timestamp,
            timeout = 1, -- Should happen quickly, but some triggered effects can take a little time
            [EVENT_TYPE.EFFECT] = onEffect
        }
        spell.extraEffect = nil -- Only set once
    end
end

local _LastSelfHeal = { event = nil, spellName = nil, targetName = nil, amount = 0, overheal = 0, absorbed = 0, isCritical = false }
local _LastSelfAura = { event = nil, spellName = nil, targetName = nil, auraType = nil, amount = 0 }
local _LastSelfDispel = { event = nil, spellName = nil, targetName = nil, exSpellId = nil, exSpellName = nil, exSpellSchool = nil }

local function onSpellSuccess(timestamp, subEvent, targetGUID, targetName, spellId, spellName, spellSchool)
    local spell = _MonitoredSpells[spellName]

    if OnCast.debug and not DEBUG_BLACKLIST[spellName] then
        OnCast.printDebugMessage(sformat("%s %s @ %0.3f - Target: %s (%s)", subEvent, spellName, timestamp, targetName or "N/A", targetGUID or ""))
    end

    if not spell then return end

    -- Remove from monitored
    _MonitoredSpells[spellName] = nil

    -- Set the timestamp
    spell.timestamp = timestamp
    spell.timeout = 6 -- Some spells have a travel time

    if EXTRA_ON_SUCCESS[spellName] then
        setExtraEffect(spell, timestamp)
    end

    -- Add the spell to succeeded spells
    _SucceededSpells[spellName] = spell

    -- Check for self heal
    if _LastSelfHeal.event and _LastSelfHeal.spellName == spellName then
        if _LastSelfHeal.timestamp == timestamp then -- Self heals happens on the same timestamp as success
            _MonitoredEvents[_LastSelfHeal.event](timestamp, _LastSelfHeal.event, "player", _LastSelfHeal.targetName, spellId, spellName, spellSchool, 
            _LastSelfHeal.amount, _LastSelfHeal.overheal, _LastSelfHeal.absorbed, _LastSelfHeal.isCritical)
        end
        _LastSelfHeal.event = nil
    end
    -- Check for self aura
    if _LastSelfAura.event and _LastSelfAura.spellName == spellName then
        if _LastSelfAura.timestamp == timestamp then-- Self aura happens on the same timestamp as success
            _MonitoredEvents[_LastSelfAura.event](timestamp, _LastSelfAura.event, "player", _LastSelfAura.targetName, spellId, spellName, spellSchool,
            _LastSelfAura.auraType, _LastSelfAura.amount)
        end
        _LastSelfAura.event = nil
    end
    -- Check for self dispell
    if _LastSelfDispel.event and _LastSelfDispel.spellName == spellName then
        if _LastSelfDispel.timestamp == timestamp then -- Self dispel happens on the same timestamp as success
            _MonitoredEvents[_LastSelfDispel.event](timestamp, _LastSelfDispel.event, "player", _LastSelfDispel.targetName, spellId, spellName, spellSchool,
            _LastSelfDispel.exSpellId, _LastSelfDispel.exSpellName, _LastSelfDispel.exSpellSchool)
        end
        _LastSelfDispel.event = nil
    end
end
_MonitoredEvents["SPELL_CAST_SUCCESS"] = onSpellSuccess


-- Transalte events into cast.event value
local TRANSLATE_EVENT = {}

-- HIT / CRIT --
-- Handle spell Damage outcomes
local function onSpellDamage(timestamp, subEvent, targetGUID, targetName, spellId, spellName, spellSchool, amount, overkill, school, resisted, blocked, absorbed, isCritical)
    local spell = _SucceededSpells[spellName]

    if OnCast.debug and not DEBUG_BLACKLIST[spellName] then
        OnCast.printDebugMessage(sformat("%s %s @ %0.3f - Target: %s (%s) - %s for %d", subEvent, spellName, timestamp, targetName or "N/A", targetGUID or "", 
        isCritical and "Crit" or "Hit", amount))
    end
    
    if not spell then return end
    
    -- Remove old spells
    if (timestamp - spell.timestamp) > spell.timeout then 
        _SucceededSpells[spellName] = nil
        return
    end    
    -- Update timestamp and timeout, allowing a small amount of time for effects or multiple hits
    spell.timestamp = timestamp
    spell.timeout = 0.01

    -- Set any extra effect
    setExtraEffect(spell, timestamp)

    local onHit = spell[EVENT_TYPE.HIT]
    if not onHit then return end
    
    -- Check if critical only
    local showHit = true
    local vars = onHit.vars
    if vars then
        showHit = showHit and ((not vars.critOnly) or isCritical)
        showHit = showHit and ((not vars.minimum) or amount >= vars.minimum)
    end
    
    if not showHit then return end

    onHit.cast.event = "DAMAGE"
    onHit.cast.outcome = isCritical and "Crit!" or "Hit"
    onHit.cast.amount = amount
    -- Override target
    if targetName then 
        onHit.cast.targetName = targetName
    end

    OnCast.outputSpellCast(onHit, onHit.cast)
end
_MonitoredEvents["SPELL_DAMAGE"] = onSpellDamage

-- Handle spell Heal outcomes
local function onSpellHeal(timestamp, subEvent, targetGUID, targetName, spellId, spellName, spellSchool, amount, overheal, absorbed, isCritical)
    local spell = _SucceededSpells[spellName]

    if OnCast.debug and not DEBUG_BLACKLIST[spellName] then
        OnCast.printDebugMessage(sformat("%s %s @ %0.3f - Target: %s (%s) - %s for %d", subEvent, spellName, timestamp, targetName or "N/A", targetGUID or "", 
        isCritical and "Crit" or "Hit", amount))
    end

    -- Check spell and timeout
    if not spell or (timestamp - spell.timestamp) > spell.timeout then 
        -- Remove old casts
        _SucceededSpells[spellName] = nil
        -- Populate last self heal for success event
        if targetGUID == PlayerGUID then
            _LastSelfHeal.event = subEvent
            _LastSelfHeal.timestamp = timestamp
            _LastSelfHeal.spellName = spellName
            _LastSelfHeal.targetName = targetName
            _LastSelfHeal.amount = amount
            _LastSelfHeal.isCritical = isCritical
        end
        return 
    end

    -- Update timestamp and timeout, allowing a small amount of time for effects or multiple hits
    spell.timestamp = timestamp
    spell.timeout = 0.01

    -- Set any extra effect
    setExtraEffect(spell, timestamp)

    local onHit = spell[EVENT_TYPE.HIT]
    if not onHit then return end
    
    -- Check if critical only
    local showHit = true
    local vars = onHit.vars
    if vars then
        showHit = showHit and ((not vars.critOnly) or isCritical)
        showHit = showHit and ((not vars.minimum) or amount >= vars.minimum)
    end
    
    if not showHit then return end

    onHit.cast.event = "HEAL"
    onHit.cast.outcome = isCritical and "Crit!" or "Hit"
    onHit.cast.amount = amount
    -- Override target
    if targetName then 
        onHit.cast.targetName = targetName
    end
    
    OnCast.outputSpellCast(onHit, onHit.cast)
end
_MonitoredEvents["SPELL_HEAL"] = onSpellHeal

-- AURA --
local TRANSLATE_AURA_TYPE = {
    ["BUFF"] = "Buff",
    ["DEBUFF"] = "Debuff"
}
local TRANSLATE_AURA_EVENT = {
    ["SPELL_AURA_APPLIED"] = {
        ["BUFF"] = "Applied",
        ["DEBUFF"] = "Inflicted"
    },
    ["SPELL_AURA_REFRESH"] = {
        ["BUFF"] = "Refreshed",
        ["DEBUFF"] = "Refreshed"
    },
    ["SPELL_AURA_APPLIED_DOSE"] = {
        ["BUFF"] = "Increased",
        ["DEBUFF"] = "Increased"
    },
}
-- Handle spell Aura outcomes
local function onSpellAura(timestamp, subEvent, targetGUID, targetName, spellId, spellName, spellSchool, auraType, amount)
    local spell = _SucceededSpells[spellName]

    if OnCast.debug and not DEBUG_BLACKLIST[spellName] then
        OnCast.printDebugMessage(sformat("%s %s @ %0.3f - Target: %s (%s) - %s x%d", subEvent, spellName, timestamp, targetName or "N/A", targetGUID or "",
        auraType, amount or 1))
    end

    -- Check spell and timeout
    if not spell or (timestamp - spell.timestamp) > spell.timeout then
        -- Remove old casts
        _SucceededSpells[spellName] = nil
        -- Populate last self aura for success event
        if targetGUID == PlayerGUID then
            _LastSelfAura.event = subEvent
            _LastSelfAura.timestamp = timestamp
            _LastSelfAura.spellName = spellName
            _LastSelfAura.targetName = targetName
            _LastSelfAura.auraType = auraType
            _LastSelfAura.amount = amount
        end
        return
    end

    -- Update timestamp and timeout, allowing a small amount of time for effects or multiple hits
    spell.timestamp = timestamp
    spell.timeout = 0.01

    -- Set any extra effect
    setExtraEffect(spell, timestamp)

    local onEffect = spell[EVENT_TYPE.EFFECT]
    if not onEffect then return end
    
    onEffect.cast.event = auraType
    onEffect.cast.outcome = TRANSLATE_AURA_EVENT[subEvent][auraType]
    onEffect.cast.amount = amount or 1
    onEffect.cast.xtype = TRANSLATE_AURA_TYPE[auraType]
    -- Override target
    if targetName then 
        onEffect.cast.targetName = targetName
    end

    OnCast.outputSpellCast(onEffect, onEffect.cast)
end
_MonitoredEvents["SPELL_AURA_APPLIED"] = onSpellAura
_MonitoredEvents["SPELL_AURA_REFRESH"] = onSpellAura
_MonitoredEvents["SPELL_AURA_APPLIED_DOSE"] = onSpellAura

-- ACTIONS -- 
local TRANSLATE_ACTION_TYPE = {
    ["SPELL_INTERRUPT"] = "Interrupted",
    ["SPELL_DISPEL"] = "Dispelled",
    ["SPELL_STOLEN"] = "Stolen",
}
TRANSLATE_EVENT["SPELL_INTERRUPT"] = "INTERRUPT"
TRANSLATE_EVENT["SPELL_DISPEL"] = "DISPEL"
TRANSLATE_EVENT["SPELL_STOLEN"] = "STEAL"
-- Handle spell Action outcomes
local function onSpellAction(timestamp, subEvent, targetGUID, targetName, spellId, spellName, spellSchool, exSpellId, exSpellName, exSpellSchool)
    local spell = _SucceededSpells[spellName]

    if OnCast.debug and not DEBUG_BLACKLIST[spellName] then
        OnCast.printDebugMessage(sformat("%s %s @ %0.3f - Target: %s (%s) - XSpell: %s", subEvent, spellName, timestamp, targetName or "N/A", targetGUID or "",
        exSpellName or "N/A"))
    end

    -- Check spell and timeout
    if not spell or (timestamp - spell.timestamp) > spell.timeout then 
        -- Remove old casts
        _SucceededSpells[spellName] = nil
        -- Populate last self dispell for success event
        if targetGUID == PlayerGUID and subEvent == "SPELL_DISPEL" then
            _LastSelfDispel.event = subEvent
            _LastSelfDispel.timestamp = timestamp
            _LastSelfDispel.spellName = spellName
            _LastSelfDispel.targetName = targetName
            _LastSelfDispel.exSpellId = exSpellId
            _LastSelfDispel.exSpellName = exSpellName
            _LastSelfDispel.exSpellSchool = exSpellSchool
        end
        return
    end
    
    -- Update timestamp and timeout, allowing a small amount of time for effects or multiple hits
    spell.timestamp = timestamp
    spell.timeout = 0.01

    -- Set any extra effect
    setExtraEffect(spell, timestamp)

    local onEffect = spell[EVENT_TYPE.EFFECT]
    if not onEffect then return end

    onEffect.cast.event = TRANSLATE_EVENT[subEvent]
    onEffect.cast.outcome = TRANSLATE_ACTION_TYPE[subEvent]
    onEffect.cast.xspell = exSpellName
    -- Override target
    if targetName then 
        onEffect.cast.targetName = targetName
    end
    
    OnCast.outputSpellCast(onEffect, onEffect.cast)
end
_MonitoredEvents["SPELL_INTERRUPT"] = onSpellAction
_MonitoredEvents["SPELL_DISPEL"] = onSpellAction
_MonitoredEvents["SPELL_STOLEN"] = onSpellAction

local function onDispelFailed(timestamp, subEvent, targetGUID, targetName, spellId, spellName, spellSchool, exSpellId, exSpellName, exSpellSchool)
    local spell = _SucceededSpells[spellName]

    if OnCast.debug and not DEBUG_BLACKLIST[spellName] then
        OnCast.printDebugMessage(sformat("%s %s @ %0.3f - Target: %s (%s) - XSpell: %s", subEvent, spellName, timestamp, targetName or "N/A", targetGUID or "",
        exSpellName or "N/A"))
    end

    -- Check spell and timeout
    if not spell or (timestamp - spell.timestamp) > spell.timeout then 
        -- Remove old casts
        _SucceededSpells[spellName] = nil
        -- Populate last self dispell for success event
        if targetGUID == PlayerGUID then
            _LastSelfDispel.event = subEvent
            _LastSelfDispel.timestamp = timestamp
            _LastSelfDispel.spellName = spellName
            _LastSelfDispel.targetName = targetName
            _LastSelfDispel.exSpellId = exSpellId
            _LastSelfDispel.exSpellName = exSpellName
            _LastSelfDispel.exSpellSchool = exSpellSchool
        end
        return
    end
    
    -- Update timestamp and timeout, allowing a small amount of time for effects or multiple hits
    spell.timestamp = timestamp
    spell.timeout = 0.01

    -- Set any extra effect
    setExtraEffect(spell, timestamp)

    local onMiss = spell[EVENT_TYPE.MISS]
    if not onMiss then return end

    onMiss.cast.event = "DISPEL FAIL"
    onMiss.cast.outcome = "Dispel Failed"
    onMiss.cast.xspell = exSpellName
    -- Override target
    if targetName then 
        onMiss.cast.targetName = targetName
    end

    OnCast.outputSpellCast(onMiss, onMiss.cast)
end
_MonitoredEvents["SPELL_DISPEL_FAILED"] = onDispelFailed

-- POWER GAIN --
local TRANSLATE_POWER_TYPE = {
    [0] = "Mana",
    [1] = "Rage",
    [2] = "Focus",
    [3] = "Energy",
}
local TRANSLATE_POWER_GAIN = {
    ["SPELL_ENERGIZE"] = "Gained",
    ["SPELL_LEECH"] = "Leeched",
    ["SPELL_DRAIN"] = "Drained",
}
TRANSLATE_EVENT["SPELL_ENERGIZE"] = "ENERGIZE"
TRANSLATE_EVENT["SPELL_LEECH"] = "LEECH"
TRANSLATE_EVENT["SPELL_DRAIN"] = "DRAIN"
-- Handle spell Power gains
local function onSpellPowerGain(timestamp, subEvent, targetGUID, targetName, spellId, spellName, spellSchool, amount, powerType)
    local spell = _SucceededSpells[spellName]
    
    if OnCast.debug and not DEBUG_BLACKLIST[spellName] then
        OnCast.printDebugMessage(sformat("%s %s @ %0.3f - Target: %s (%s) - %s %d", subEvent, spellName, timestamp, targetName or "N/A", targetGUID or "",
        TRANSLATE_POWER_TYPE[powerType], amount))
    end

    if not spell then return end

    -- Remove old spells
    if (timestamp - spell.timestamp) > spell.timeout then 
        _SucceededSpells[spellName] = nil
        return
    end
    -- Update timestamp and timeout, allowing a small amount of time for effects or multiple hits
    spell.timestamp = timestamp
    spell.timeout = 0.01

    -- Set any extra effect
    setExtraEffect(spell, timestamp)

    local onEffect = spell[EVENT_TYPE.EFFECT]
    if not onEffect then return end

    onEffect.cast.event = TRANSLATE_EVENT[subEvent]
    onEffect.cast.outcome = TRANSLATE_POWER_GAIN[subEvent]
    onEffect.cast.amount = amount
    onEffect.cast.xtype = TRANSLATE_POWER_TYPE[powerType]
    -- Override target
    if targetName then 
        onEffect.cast.targetName = targetName
    end
    
    OnCast.outputSpellCast(onEffect, onEffect.cast)
end
_MonitoredEvents["SPELL_LEECH"] = onSpellPowerGain
_MonitoredEvents["SPELL_DRAIN"] = onSpellPowerGain
-- Translate args for energize
local function onSpellEnergize(timestamp, subEvent, targetGUID, targetName, spellId, spellName, spellSchool, amount, overEnergize, powerType)
    onSpellPowerGain(timestamp, subEvent, targetGUID, targetName, spellId, spellName, spellSchool, amount, powerType)
end
_MonitoredEvents["SPELL_ENERGIZE"] = onSpellEnergize

-- GENERAL OUTCOME --
local TRANSLATE_OUTCOME = {
    ["SPELL_CREATE"] = "Created",
    ["SPELL_SUMMON"] = "Summoned",
    ["SPELL_RESURRECT"] = "Resurrected",
}
TRANSLATE_EVENT["SPELL_CREATE"] = "CREATE"
TRANSLATE_EVENT["SPELL_SUMMON"] = "SUMMON"
TRANSLATE_EVENT["SPELL_RESURRECT"] = "RESURRECT"
-- Handle other spell outcomes; e.g. Create, Summon, Ressurrect
local function onSpellOutcome(timestamp, subEvent, targetGUID, targetName, spellId, spellName, spellSchool)
    local spell = _SucceededSpells[spellName]

    if OnCast.debug and not DEBUG_BLACKLIST[spellName] then
        OnCast.printDebugMessage(sformat("%s %s @ %0.3f - Target: %s (%s)", subEvent, spellName, timestamp, targetName or "N/A", targetGUID or ""))
    end

    if not spell then return end

    -- Remove old spells
    if (timestamp - spell.timestamp) > spell.timeout then 
        _SucceededSpells[spellName] = nil
        return
    end
    -- Update timestamp and timeout, allowing a small amount of time for effects or multiple hits
    spell.timestamp = timestamp
    spell.timeout = 0.01

    -- Set any extra effect
    setExtraEffect(spell, timestamp)

    local onEffect = spell[EVENT_TYPE.EFFECT]
    if not onEffect then return end

    onEffect.cast.event = TRANSLATE_EVENT[subEvent]
    onEffect.cast.outcome = TRANSLATE_OUTCOME[subEvent]
    -- Override target
    if targetName then 
        onEffect.cast.targetName = targetName
    end

    OnCast.outputSpellCast(onEffect, onEffect.cast)
end
_MonitoredEvents["SPELL_CREATE"] = onSpellOutcome
_MonitoredEvents["SPELL_SUMMON"] = onSpellOutcome
_MonitoredEvents["SPELL_RESURRECT"] = onSpellOutcome

-- MISS --
local TRANSLATE_MISS_TYPE = {
    ["ABSORB"] = "Absorbed",
    ["BLOCK"] = "Blocked",
    ["DEFLECT"] = "Deflected",
    ["DODGE"] = "Dodged",
    ["EVADE"] = "Evaded",
    ["IMMUNE"] = "Immune",
    ["MISS"] = "Missed",
    ["PARRY"] = "Parried",
    ["REFLECT"] = "Reflected",
    ["RESIST"] = "Resisted",
}
-- Handle spell Miss outcomes
local function onSpellMiss(timestamp, subEvent, targetGUID, targetName, spellId, spellName, spellSchool, missType)
    local spell = _SucceededSpells[spellName]

    if OnCast.debug and not DEBUG_BLACKLIST[spellName] then
        OnCast.printDebugMessage(sformat("%s %s @ %0.3f - Target: %s (%s) - %s", subEvent, spellName, timestamp, targetName or "N/A", targetGUID or "",
        missType))
    end

    if not spell then return end

    -- Remove old spells
    if (timestamp - spell.timestamp) > spell.timeout then
        _SucceededSpells[spellName] = nil
        return
    end

    -- Update timestamp and timeout, allowing a small amount of time for effects or multiple hits
    spell.timestamp = timestamp
    spell.timeout = 0.01
    
    local onMiss = spell[EVENT_TYPE.MISS]
    if not onMiss then return end

    onMiss.cast.event = missType
    onMiss.cast.outcome = TRANSLATE_MISS_TYPE[missType]
    onMiss.cast.reason = TRANSLATE_MISS_TYPE[missType]
    -- Override target
    if targetName then 
        onMiss.cast.targetName = targetName
    end

    OnCast.outputSpellCast(onMiss, onMiss.cast)
end
_MonitoredEvents["SPELL_MISSED"] = onSpellMiss


-- SPELL_CAST_FAILED
local function getOnCooldown(spellName, spellId, originalReason)
    local start, duration, enabled = WOW.GetSpellCooldown(spellId)
    -- Check if already active, e.g. Combustion, Inner Focus
    if enabled == 0 then
        return "Already Active"
    end

    -- Check cooldown
    if (start > 0 and duration > 0) then
        local time = WOW.GetTime()
        local delta = start + duration - time
        return "On Cooldown", delta, "cooldown"
    end
    return originalReason
end

local function getLossOfControl(spellName, spellId, originalReason, reasonText)
    --local n = C_LossOfControl.GetActiveLossOfControlDataCount()
    --print(sformat("Loss Of Control (x%d):", n))
    --for i=1, n do
      --  local locData = C_LossOfControl.GetActiveLossOfControlData(i)
      --  print(locData.locType, locData.spellID, locData.displayText, locData.startTime, locData.duration)
    --end
    -- Check for Loss of Control; e.g. Stunned, Silenced, Feared, etc.
    local start, duration = WOW.GetSpellLossOfControlCooldown(spellId)
    if (start > 0 and duration > 0) then
        local time = WOW.GetTime()
        local delta = start + duration - time
        return reasonText, delta, "lossofcontrol"
    end
    return reasonText.." {Mechanic}"
    --return originalReason
end

local SP_FAILED_PBM = SPELL_FAILED_PREVENTED_BY_MECHANIC
-- Translate fail string into something more meaningful
local FAIL_REASON_FUNCTIONS = {
    [SPELL_FAILED_NOT_READY] = getOnCooldown,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_BANISH))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Banished") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_CHARM))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Charmed") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_CONFUSE))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Confused") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_CYCLONE))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Cycloned") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_DAZE))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Dazed") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_DISARM))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Disarmed") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_DISORIENT))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Disoriented") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_DISTRACT))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Distracted") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_FEAR))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Feared") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_FREEZE))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Frozen") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_HORROR))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Horrified") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_INCAPACITATE))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Incapacitated") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_INVULNERABILITY))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Invulnerable") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_PACIFY))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Pacified") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_PACIFYSILENCE))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Disabled") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_POLYMORPH))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Polymorphed") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_POSSESS))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Possessed") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_ROOT))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Rooted") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_SAP))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Sapped") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_SHACKLE_UNDEAD))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Shackled") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_SILENCE))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Silenced") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_SLEEP))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Asleep") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_SNARE))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Snared") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_STUN))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Stunned") end,
    [sformat(SP_FAILED_PBM,slower(LOSS_OF_CONTROL_DISPLAY_TAUNT))] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Taunted") end,
    [SPELL_FAILED_FLEEING] = function(spellName, spellId, origReason) return getLossOfControl(spellName, spellId, origReason, "Fleeing") end,
    [SPELL_FAILED_MOVING] = function() return "Movement" end,
    [SPELL_FAILED_LINE_OF_SIGHT] = function() return "Line Of Sight" end,
}

-- Keep track of failed casts by spell name (Classic doesn't give spellId)
local _LastFailureReason = {}
function ExtEvent.GetLastFailureReason(spellName, spellId)
    local reason, delta, type = _LastFailureReason[spellName], nil, nil
    local reasonFunc = FAIL_REASON_FUNCTIONS[reason]
    if reasonFunc then
        reason, delta, type = reasonFunc(spellName, spellId, reason)
    end
    return reason, delta, type
end

local function onSpellCastFailed(timestame, subEvent, targetGUID, targetName, spellId, spellName, spellSchool, failType)
    _LastFailureReason[spellName] = failType
end
_MonitoredEvents["SPELL_CAST_FAILED"] = onSpellCastFailed

function ExtEvent.DebugLists()
    OnCast.printInfoMessage("Monitored spells:")
    for k,_ in pairs(_MonitoredSpells) do
        OnCast.printInfoMessage(sformat(" - %s", k))
    end
    OnCast.printInfoMessage("Succeeded Spells:")
    local time = time()
    for k,v in pairs(_SucceededSpells) do
        OnCast.printInfoMessage(sformat(" - %s -Age: %0.2fs",k,time - v.timestamp))
    end
end
