local _, ns = ...
local OnCast = ns.addon
local MacroConditionals = OnCast:NewModule("MacroConditionals")

-- Lua API
local sformat = string.format
local smatch = string.match
local slower = string.lower

-- WoW API
local WOW = OnCast.WOWAPI or {}
WOW.GetMouseButtonClicked = GetMouseButtonClicked -- https://wowpedia.fandom.com/wiki/API_GetMouseButtonClicked
WOW.IsModifierKeyDown = IsModifierKeyDown -- https://wowpedia.fandom.com/wiki/API_IsModifierKeyDown
WOW.UnitExists = UnitExists -- https://wow.gamepedia.com/API_UnitExists
WOW.UnitIsEnemy = UnitIsEnemy -- https://wowpedia.fandom.com/wiki/API_UnitIsEnemy
WOW.UnitIsFriend = UnitIsFriend -- https://wow.gamepedia.com/API_UnitIsFriend
WOW.UnitIsDeadOrGhost = UnitIsDeadOrGhost -- https://wowpedia.fandom.com/wiki/API_UnitIsDeadOrGhost
WOW.GetUnitName = GetUnitName -- https://wow.gamepedia.com/API_GetUnitName

-- Values
local playerName = WOW.GetUnitName("player", true)

-- Modules
local Util = nil

-- Constants
local CONDITION_SPLIT = ','
local CONDITION_VALUE_SPLIT = ':'
local VALUE_SPLIT = '/'

-- Module functions
function MacroConditionals:OnInitialize()
    Util = OnCast:GetModule("Util")

    Util.addSplitPattern(CONDITION_SPLIT)
    Util.addSplitPattern(CONDITION_VALUE_SPLIT)
    Util.addSplitPattern(VALUE_SPLIT)
end

function MacroConditionals:OnEnable()
end

function MacroConditionals:OnDisable()
end

-- Map of Mouse Button number to name
local BUTTON_NAMES = {}
BUTTON_NAMES["1"] = "LeftButton"
BUTTON_NAMES["2"] = "RightButton"
BUTTON_NAMES["3"] = "MiddleButton"
BUTTON_NAMES["4"] = "Button4"
BUTTON_NAMES["5"] = "Button5"

local function evaluateButton(value)
    local button = WOW.GetMouseButtonClicked()
    -- Button 1 is considered to be pressed if a hotkey was used
    if (not button) then return (value == "1") end 
    
    local buttonName = (value and BUTTON_NAMES[value])
    if buttonName then
        return (button == buttonName)
    else
        OnCast.printErrorMessage(sformat("Unknown button value: %s", value or "<nil>"))
    end
    return false
end

-- Shift / Crtl / Alt Key Modifier functions
local MODKEY_FUNCTIONS = {}
MODKEY_FUNCTIONS["shift"] = IsShiftKeyDown
MODKEY_FUNCTIONS["s"] = MODKEY_FUNCTIONS["shift"]
MODKEY_FUNCTIONS["lshift"] = IsLeftShiftKeyDown
MODKEY_FUNCTIONS["ls"] = MODKEY_FUNCTIONS["lshift"]
MODKEY_FUNCTIONS["rshift"] = IsRightShiftKeyDown
MODKEY_FUNCTIONS["rs"] = MODKEY_FUNCTIONS["rshift"]
MODKEY_FUNCTIONS["ctrl"] = IsControlKeyDown
MODKEY_FUNCTIONS["c"] = MODKEY_FUNCTIONS["ctrl"]
MODKEY_FUNCTIONS["lctrl"] = IsLeftControlKeyDown
MODKEY_FUNCTIONS["lc"] = MODKEY_FUNCTIONS["lctrl"]
MODKEY_FUNCTIONS["rctrl"] = IsRightControlKeyDown
MODKEY_FUNCTIONS["rc"] = MODKEY_FUNCTIONS["rctrl"]
MODKEY_FUNCTIONS["alt"] = IsAltKeyDown
MODKEY_FUNCTIONS["a"] = MODKEY_FUNCTIONS["alt"]
MODKEY_FUNCTIONS["lalt"] = IsLeftAltKeyDown
MODKEY_FUNCTIONS["la"] = MODKEY_FUNCTIONS["lalt"]
MODKEY_FUNCTIONS["ralt"] = IsRightAltKeyDown
MODKEY_FUNCTIONS["ra"] = MODKEY_FUNCTIONS["ralt"]

local _tempModKeys = {}
local function evaluateModKeys(value)
    if not value then return WOW.IsModifierKeyDown() end
    local n = Util.stringSplitEF(value, VALUE_SPLIT, _tempModKeys)
    for i=1, n do
        local modKey = _tempModKeys[i]
        local keyFunc = MODKEY_FUNCTIONS[modKey]
        if keyFunc then
            if keyFunc() then return true end -- OR Condition
        else
            OnCast.printErrorMessage(sformat("Unknown modifier key: %s", modKey))
        end
    end
    return false
end

local MODIFIER_CONDITION_FUNCS = {}
-- Buttons
MODIFIER_CONDITION_FUNCS["button"] = evaluateButton
MODIFIER_CONDITION_FUNCS["btn"] = MODIFIER_CONDITION_FUNCS["button"]
MODIFIER_CONDITION_FUNCS["b"] = MODIFIER_CONDITION_FUNCS["button"]
MODIFIER_CONDITION_FUNCS["nobutton"] = function(value) return not evaluateButton(value) end
MODIFIER_CONDITION_FUNCS["nobtn"] = MODIFIER_CONDITION_FUNCS["nobutton"]
MODIFIER_CONDITION_FUNCS["nb"] = MODIFIER_CONDITION_FUNCS["nobutton"]
MODIFIER_CONDITION_FUNCS["modifier"] = evaluateModKeys
MODIFIER_CONDITION_FUNCS["mod"] = MODIFIER_CONDITION_FUNCS["modifier"]
MODIFIER_CONDITION_FUNCS["m"] = MODIFIER_CONDITION_FUNCS["modifier"]
MODIFIER_CONDITION_FUNCS["nomodifier"] = function(value) return not evaluateModKeys(value) end
MODIFIER_CONDITION_FUNCS["nomod"] = MODIFIER_CONDITION_FUNCS["nomodifier"]
MODIFIER_CONDITION_FUNCS["nm"] = MODIFIER_CONDITION_FUNCS["nomodifier"]

-- Target
local UNIT_CONDITION_FUNCS = {}
UNIT_CONDITION_FUNCS["exists"] = function(unitId) return WOW.UnitExists(unitId) end
UNIT_CONDITION_FUNCS["harm"] = function(unitId) return WOW.UnitIsEnemy("player",unitId) end
UNIT_CONDITION_FUNCS["noharm"] = function(unitId) return not WOW.UnitIsEnemy("player",unitId) end
UNIT_CONDITION_FUNCS["help"] = function(unitId) return WOW.UnitIsFriend("player", unitId) end
UNIT_CONDITION_FUNCS["nohelp"] = function(unitId) return not WOW.UnitIsFriend("player", unitId) end
UNIT_CONDITION_FUNCS["dead"] = function(unitId) return WOW.UnitIsDeadOrGhost(unitId) end
UNIT_CONDITION_FUNCS["nodead"] = function(unitId) return not WOW.UnitIsDeadOrGhost(unitId) end

-- Target Aliases
local TARGET_ALIAS = {}
TARGET_ALIAS["p"] = "player"
TARGET_ALIAS["t"] = "target"
TARGET_ALIAS["mo"] = "mouseover"
TARGET_ALIAS["f"] = "focus"
TARGET_ALIAS["tt"] = "targettarget"

local _tempConditionValues = {}
local _tempConditionValue = {}
function MacroConditionals.Evaluate(conditionStr)
    local n = Util.stringSplitEF(conditionStr, CONDITION_SPLIT, _tempConditionValues)
    if n < 1 then return true, nil end
    
    -- Check for unitId
    local unitId = smatch(_tempConditionValues[1], "^@(%w+)")
    unitId = unitId and TARGET_ALIAS[unitId] or unitId
    
    -- Returned dynamic units
    local dynamicUnits = nil
    if not unitId then
        dynamicUnits = {}
        dynamicUnits["player"] = playerName
        if WOW.UnitExists("target") then dynamicUnits["target"] = WOW.GetUnitName("target",true) end
        if WOW.UnitExists("mouseover") then dynamicUnits["mouseover"] = WOW.GetUnitName("mouseover",true) end
        if WOW.UnitExists("focus") then dynamicUnits["focus"] = WOW.GetUnitName("focus",true) end
        if WOW.UnitExists("targettarget") then dynamicUnits["targettarget"] = WOW.GetUnitName("targettarget",true) end
    end

    -- Loop conditions
    for i=(unitId and 2 or 1), n do
        local conditionValue = _tempConditionValues[i]
        _tempConditionValue[1] = nil
        _tempConditionValue[2] = nil
        
        Util.stringSplitEF(conditionValue, CONDITION_VALUE_SPLIT, _tempConditionValue)
        
        local condition = _tempConditionValue[1]
        local value = _tempConditionValue[2]
        
        local conditionFunction = MODIFIER_CONDITION_FUNCS[condition]
        if conditionFunction then
            local ok = conditionFunction(value)
            if (not ok) then return false end -- AND Condition
        else
            conditionFunction = UNIT_CONDITION_FUNCS[condition]
            if conditionFunction then
                if unitId then
                    local ok = conditionFunction(unitId)
                    if (not ok) then return false end -- AND Condition
                else -- Evaluate dynamic units
                    local ok = false
                    
                    if dynamicUnits["player"] and conditionFunction("player") then ok = true
                    else dynamicUnits["player"] = nil end

                    if dynamicUnits["target"] and conditionFunction("target") then ok = true
                    else dynamicUnits["target"] = nil end

                    if dynamicUnits["mouseover"] and conditionFunction("mouseover") then ok = true
                    else dynamicUnits["mouseover"] = nil end

                    if dynamicUnits["focus"] and conditionFunction("focus") then ok = true
                    else dynamicUnits["focus"] = nil end

                    if dynamicUnits["targettarget"] and conditionFunction("targettarget") then ok = true
                    else dynamicUnits["targettarget"] = nil end

                    if (not ok) then return false end -- AND Condition, no dynamic units
                end
            else
                OnCast.printErrorMessage(sformat("Unknown Condition: %s", condition))
                return false
            end
        end
    end
    return true, unitId, dynamicUnits
end

function MacroConditionals.EvaluateUnitConditions(conditions, unitId)
    for i=1, #conditions do
        local condition = conditions[i]
        local conditionFunction = UNIT_CONDITION_FUNCS[condition]
        if conditionFunction then
            local ok = conditionFunction(unitId)
            if (not ok) then return false end -- AND Condition
        else
            OnCast.printErrorMessage(sformat("Unknown Unit Condition: %s", condition))
        end
    end
    return true
end

function MacroConditionals.Help()
    OnCast.printInfoMessage("-- Conditions --", OnCast.HEADING_COLOR)
    OnCast.printInfoMessage("All lower case, surrounded by [ ] and separated by a , (comma) with no spaces.")
    OnCast.printInfoMessage("Example: [@target,nobtn:2,mod:alt/shift]")
    OnCast.printInfoMessage("@unit: Must be first. A valid macro unit id; e.g. @target @mouseover")
    OnCast.printInfoMessage("Mouse Button: [btn:n] or [nobtn:n] where n is a number from 1 to 5")
    OnCast.printInfoMessage("Modifier Keys: [mod] for any, [mod:key] or [nomod] for none, [nomod:key]")
    OnCast.printInfoMessage("- Keys: shift, rshift, lshift, ctrl, rctrl, lctrl, alt, ralt, lalt")
    OnCast.printInfoMessage("- Combine keys with / (forward slash); e.g. shift/ctrl")
    OnCast.printInfoMessage(sformat("Unit State: %s", slower(Util.concatTableKeys(UNIT_CONDITION_FUNCS, ", "))))
end
