local _, ns = ...
local OnCast = ns.addon
local module = OnCast:NewModule("ExtParrot")

-- Extended channel name
local CHANNEL = "PARROT"

-- Lua APIs
local sformat = string.format
local supper = string.upper
local sgsub = string.gsub
local tonumber = tonumber
local mfloor = math.floor

-- Modules
local Util = nil

-- Upvalue to Parrot
local _parrot = nil

function module:OnInitialize()
    Util = OnCast:GetModule("Util")
end

function module:OnEnable()
    _parrot = Parrot
    if _parrot then
        OnCast.AddExtendedChannel(CHANNEL,module)
    end
end

function module:OnDisable()
    OnCast.RemoveExtendedChannel(CHANNEL)
    _parrot = nil
end

-- Sticky 'true' values
local STICKY_VALUES = {
    ["STICKY"] = true,
    ["S"] = true,
    ["TRUE"] = true,
    ["T"] = true,
}

-- Valid outline values
local OUTLINE_VALUES = {
    ["NONE"] = "None",
    ["THIN"] = "OUTLINE",
    ["THICK"] = "THICKOUTLINE",
    ["MONOCHOROME"] = "MONOCHROME",
    ["MONO"] = "MONOCHROME",
    ["THINMONO"] = "OUTLINE,MONOCHOROME",
}

function module.Help()
    OnCast.printInfoMessage("-- PARROT --", OnCast.HEADING_COLOR)
    OnCast.printInfoMessage("Usage: parrot(Scroll_Area;Sticky;Color;Size;Outline)")    
    OnCast.printInfoMessage(sformat("Scroll_Area: %s", Util.concatTableKeys(_parrot:GetScrollAreasChoices(), ", ", function(key) return sformat("'%s'", sgsub(key, ' ', '_')) end)))
    OnCast.printInfoMessage("Sticky: True / False")
    OnCast.printInfoMessage("Color: RRGGBB in Hexidecimal; e.g. FFC4A9")
    OnCast.printInfoMessage("Size: Number; e.g. 16")
    OnCast.printInfoMessage("Outline: None, Thin, Thick, Mono, ThinMono")
    OnCast.printInfoMessage("Use _ (underscore) for default value: parrot(area;_;_;24;thin)")
end

-- Check if a var is not blank; i.e. _ (underscore)
local function notBlank(value)
    return (value and value ~= "_")
end

function module.VarsToTable(vars, n)
    -- Required: Area
    local area = vars[1]
    if not area then
        OnCast.printErrorMessage(sformat("%s Scroll Area Required!", CHANNEL))
        return false, nil
    end
    local scrollAreas = _parrot:GetScrollAreasChoices()
    if not scrollAreas[area] then
        area = sgsub(area, '_', ' ')
        if not scrollAreas[area] then
            OnCast.printErrorMessage(sformat("Invalid %s Scroll Area '%s'.\nExpected one of: %s", CHANNEL, area,
            Util.concatTableKeys(scrollAreas, ", ", function(key) return sformat("'%s'", key) end)))
            return false, nil
        end
    end
    
    vars.area = area
    vars[1] = nil

    -- Sticky
    local sticky = vars[2]
    if notBlank(sticky) then
        vars.sticky = STICKY_VALUES[supper(sticky)]
    end
    vars[2] = nil

    -- Early exit for common use case
    if n <= 2 then return true, vars end

    -- Color
    local color = vars[3]
    if notBlank(color) then
        if #color ~= 6 then
            OnCast.printErrorMessage("Invalid color. Expected format: RRGGBB")
            return false, nil
        end
        local nc, colors = Util.colorStringToTable(color)
        if nc == 3 then
            vars.r = Util.hexToInt(colors[1]) / 255
            vars.g = Util.hexToInt(colors[2]) / 255
            vars.b = Util.hexToInt(colors[3]) / 255
        else
            OnCast.printErrorMessage("Invalid color. Expected Hexidecimal digits [0-9,A-F]")
            return false, nil
        end
    end
    vars[3] = nil

    -- Size
    local size = vars[4]
    if notBlank(size) then
        local size_n = tonumber(size)
        if size_n then
            vars.size = size_n
        else
            OnCast.printErrorMessage(sformat("Invalid %s Font Size '%s'. Expected number.", CHANNEL, size))
            return false, nil
        end
    end
    vars[4] = nil

    -- Outline
    local outline = vars[5]
    if notBlank(outline) then
        vars.outline = OUTLINE_VALUES[supper(outline)]
        if not vars.outline then
            OnCast.printErrorMessage(sformat("Invalid %s Outline value '%s'.\nExpected one of: %s", CHANNEL, outline, 
            Util.concatTableKeys(OUTLINE_VALUES, ", ")))
            return false, nil
        end
    end
    vars[5] = nil

    -- Clear any other values
    for i=6, n do vars[i] = nil end

    return true, vars
end

-- Translate r,g,b 0>1 values into hex
local function getColorFromRGB(r,g,b)
    if not (r or g or b) then return nil end
    local colors = {}
    colors[1] = Util.intToHex(mfloor((r or 1) * 255 + 0.5))
    colors[2] = Util.intToHex(mfloor((g or 1) * 255 + 0.5))
    colors[3] = Util.intToHex(mfloor((b or 1) * 255 + 0.5))
    return Util.colorTableToString(colors)
end

function module.VarsToArray(vars)
    local array = {}
    array[1] = vars.area or "_"
    array[2] = vars.sticky and "Sticky" or "Normal"
    array[3] = getColorFromRGB(vars.r, vars.g, vars.b) or "_"
    array[4] = vars.size or "_"
    array[5] = vars.outline or "_"
    return array
end

function module.ShowMessage(message, vars, cast)
    if _parrot then
        local v = vars or {}
         -- text, area, sticky, r, g, b, font, fontSize, outline, icon
        _parrot:ShowMessage(message, v.area, v.sticky, vars.r, vars.g, vars.b, nil, v.size, v.outline, cast and cast.spell.icon)
    end
end
